from typing import Any


class Entry:
    """Base class of all entry types."""

    __slots__ = ("implementation", "service")

    @property
    def key(self) -> str:
        return self.service.__name__

    def __init__(self, implementation: type, service: type = None) -> None:
        if not isinstance(implementation, type):
            raise ValueError("implementation has to be a type")
        if service is None:
            service = implementation
        if not isinstance(service, type):
            raise ValueError("service has to be a type")

        self.implementation = implementation
        self.service = service



class InstanceEntry(Entry):
    """Returns a preconstructed instance."""

    __slots__ = ("arguments", "keywords")

    def __init__(self, implementation: type, service: type = None, arguments: tuple[Any] = None, keywords: dict[str, Any] = None) -> None:
        super().__init__(implementation, service)
        self.arguments = arguments if arguments is not None else ()
        self.keywords = keywords if keywords is not None else {}


class SingletonEntry(Entry):
    """A singlton entry. Returns always the same instance of an object."""

    __slots__ = ("instance",)

    def __init__(self, implementation: type, service: type = None, instance: object = None) -> None:
        super().__init__(implementation, service)
        self.instance = instance



from typing import Any, Type

from .container import Container, T

_container: Container = None


def set_container(container: Container) -> None:
    global _container
    if not isinstance(container, Container):
        raise Exception("container has to be type of Container")
    _container = container


def register(implementation: type, service: type = None) -> None:
    _container.register(implementation, service)


def register_instance(
    implementation: type, service: type = None, args: tuple = (), kwds: dict[str, Any] = None
) -> None:
    _container.register_instance(implementation, service, args, kwds)


def register_singleton(implementation: type, service: type = None, instance: object = None) -> None:
    _container.register_singleton(implementation, service, instance)


def get(type: Type[T], default: Any = None) -> T:
    instance = _container.get(type)
    if instance is None:
        return default
    return instance



import math
from typing import Any, Callable

from colorama import Fore, Style


class ProgressBar:
    __slots__ = ("value", "total", "elements","title", "digits", "percentage")

    MAX_SIZE = 50

    def __init__(
        self,
        elements: int,
        title: str = None,
        action: Callable[..., Any] = None,
        percentage: bool = True,
    ) -> None:
        self.value = 0
        self.total = elements
        self.elements = None
        if not isinstance(elements, int):
            self.total = len(elements)
            self.elements = elements

        self.title = title
        if self.title is None:
            self.title = "Progress"

        self.digits = len(str(self.total))
        self.percentage = percentage
        print(f"  {self.title}")

        if action is not None:
            self._run(action)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if value is not None:
            self._increment(True)
            print()

    def _run(self, action: Callable[..., Any]) -> None:
        collection = range(self.total)
        if self.elements is not None:
            collection = self.elements

        for x in collection:
            action(x)
            self._increment()

    def increment(self) -> None:
        self._increment()

    def _increment(self, cancel: bool = False) -> None:
        self.value += 1
        color = ""
        if cancel:
            self.value -= 1
            color = f"{Fore.RED}"
        elif self.value == self.total:
            self.value = self.total
            color = f"{Fore.GREEN}"

        scale = ProgressBar.MAX_SIZE / self.total
        progress = scale * self.value

        full_blocks = math.floor(progress)
        subs = int((progress - math.floor(progress)) * 8)
        sub_block = chr(0x2590 - subs)
        if subs == 0:
            sub_block = " "
            if self.value == self.total:
                sub_block = ""

        text = f"  {color}{chr(0x2588) * full_blocks}{sub_block}{' ' * (ProgressBar.MAX_SIZE - full_blocks - 1)}{chr(0x258F)}"
        if self.percentage:
            text += f"{int(self.value * 100 / self.total) :>3} %{Style.RESET_ALL}"
        else:
            text += f"{str.rjust(str(self.value), self.digits)}/{self.total}{Style.RESET_ALL}"

        print(text, end="\r")

        if self.total == self.value:
            print()


# for i in range(1, 11):
#     print(f"{i=}, ", end="")
#     full_blocks = i // 8
#     print(f"{full_blocks=}, ", end="")
#     sub_block = i % 8
#     if sub_block == 0:
#         sub_block = 8
#     # sub_block = 8 if sub_block == 0 else sub_block
#     print(f"{sub_block=}, ", end="")
#     print(chr(0x2590 - sub_block))



import threading


class Event:
    """Base class for all custom events"""

    @property
    def name(self) -> str:
        return type(self).__name__

    def __str__(self) -> str:
        attributes = vars(self).items()
        out = self.name
        if attributes:
            out +=  f" {', '.join([f'{k}({v})' for k, v in attributes])}"
        return out

    def __repr__(self) -> str:
        return str(self)


class EventManager:
    __slots__ = ("_subs", "_entries")

    def __init__(self) -> None:
        self._subs: dict[int, object] = {}
        self._entries: dict[int, list[type]] = {}

    def subscribe(self, subscriber: object) -> None:
        self._subscribe(subscriber)

    def _subscribe(self, subscriber: object) -> int:
        sid = id(subscriber)
        if not sid in self._subs:
            self._subs[sid] = subscriber
        return sid

    def unsubscribe(self, subscriber: object) -> None:
        sid = id(subscriber)
        if not sid in self._subs:
            return
        del self._subs[sid]
        if sid in self._entries:
            del self._entries[sid]

    def register_for(self, subscriber: object, EventType: type) -> None:
        sid = self._subscribe(subscriber)
        if not sid in self._entries:
            self._entries[sid] = []
        self._entries[sid].append(EventType)

    def unregister(self, subscriber: object, EventType: type) -> None:
        sid = id(subscriber)
        if not sid in self._entries:
            return
        if EventType in self._entries[sid]:
            self._entries[sid].remove(EventType)

    def publish(self, sender, event: Event) -> None:
        subs = self._subs.copy()
        for sid in subs:
            if not sid in self._entries:
                continue
            if not type(event) in self._entries[sid]:
                continue
            getattr(subs[sid], f"on_{ event.name.lower() }")(event, sender)

    def publish_async(self, sender, event: Event) -> None:
        threading.Thread(target=self.publish, args=(event, sender), daemon=True).start()
